import config
from util import *
from domain_model import *

import os
import stat

class Event:
    def __repr__(self):
        return "<" + self.__class__.__name__ + ">"

class OpenEvent (Event):
    def handle(self, start_process, responsible_process, fields):
        #print fields            responsible_process.add_transition('activate', link_name)
        rname = the_resource_name(fields)
        resource_sc = the_resource_sc(fields)
        if is_directory(fields):
            dirobj = resource_factory(PRGM, rname, 'Dir', resource_sc)
            responsible_process.add_transition('open', rname)
        elif is_file(fields):
            fileobj = resource_factory(PRGM, rname, 'File', resource_sc)
            fd = the_file_descriptor(fields) # may not need this yet
            if  readerp(fields):
                responsible_process.add_transition('read', rname)
            elif writerp(fields):
                 responsible_process.add_transition('write', rname)

class PipeEvent (Event):
    def handle(self, start_process,responsible_process, fields):
        #need to set up read or write events between responsible_process and process at other end of pipe
        #each pipe statement sets up two fd's
        #actions on either constitute R/W events
##        print fields
        fd2 = fields[-2]
        fd1 = the_resource_name(fields)
        pwp = pipe_writerp(fields)
        prp = pipe_readerp(fields)
        if pwp or prp:
            direct = switch(fd2, fd1)
            reverse = switch(fd1, fd2)
            if direct in resources_dict.keys():
                pipe = resources_dict[direct]
            elif reverse in resources_dict.keys():
                pipe = resources_dict[reverse]
            else:
                pipe = Fifo_file(direct,direct, responsible_process, [fd1, fd2])
            r_or_w = 'write'
            if prp:
                r_or_w = 'read'
            responsible_process.add_transition(r_or_w, pipe.name)

def switch (fd1, fd2):
    return fd2+"&"+fd1

class CloneorShareEvent(Event):
    def handle (self,start_process, responsible_process, fields):
             #create a new process
            new_pid = the_resource_name(fields)
            new_process = resource_factory(PRGM, responsible_process.name + "_" + str(new_pid), 'Process',  the_process_sc(fields))
            start_process.processes[new_pid] = new_process
            new_process.pid = new_pid
            link_name = responsible_process.name +'-execs-' + new_process.name
            spawn_link = resource_factory(PRGM, link_name, 'Spawn', responsible_process.security_context)
            responsible_process.add_transition('activate', link_name)
            new_process.add_transition('spawn', link_name)

class CloneEvent(CloneorShareEvent):
    share = 0

class ShareEvent(CloneorShareEvent):
    share = 1

class ExecEvent(Event):
    def handle (self, start_process,responsible_process, fields):
        #print fields
        ok = succeeds(fields)
        #print ok
        if ok: # exec works
            #print 'wins'
            #create a read event for the responsible process
            filestr = the_resource_name(fields)
            #print filestr
            #file gets its sc from the line; process gets its sc by removing exec_ from files sc
            file_sc = fields[3]
            file = resource_factory(PRGM, filestr, 'File', file_sc) # might have _exec_t in it
            if file_sc.find('polgen_temp') != -1:
                process_sc =  fields[3][0:-7] + fields[3][-2:]
            else:
                process_sc = 'unknown'
                #print process_name(filestr)
                
            process = resource_factory(PRGM, process_name(filestr), 'Process', process_sc) # was the_resource_sc(fields))
            process.assoc_binary_file = file
            start_process.processes[fields[0]] = process
            responsible_process.executing.append(process)
            process.executed_by.append(responsible_process)
            responsible_process.add_transition('exec', filestr)
            link_name = responsible_process.name + '-directly-execs-' + process.name
            spawn_link = resource_factory(PRGM, link_name, 'Spawn', responsible_process.security_context)
            responsible_process.add_transition('activate', link_name)
            process.add_transition('spawn', link_name)
            #print responsible_process.name
            #print responsible_process.out_edges
        else:
            #print "exec that fails"
            #print fields
            start_process.failures.append(fields)

class StreamEvent(Event):
    def handle (self, start_process,responsible_process, fields):
        return 'Not handling stream events yet'

class MiscEvent(Event):
    def handle (self, start_process,responsible_process, fields):
        responsible_process.special_calls.append(fields[2])
        
    
class UnlinkEvent(Event):
    def handle (self, start_process,responsible_process, fields):
        return 'Not handling unlink events yet'

class SockPrEvent (Event):
    def handle (self, start_process,responsible_process, fields):
        return 'Not handling sockpr events yet'
    
class SocketEvent(Event):
    #pid|Socket|operation|parent_tracker|addr_info|R or ""|??? or ""|W or ""|??? or ""|fd|tracker_id|security_context
    # 0    1        2          3            4        5        6         7       8       9   10          11
    def handle (self,start_process, responsible_process, fields):
        operation = fields[2]
        port= the_port(fields)
        tracker = the_tracker_nbr(fields)
        id = port
        if operation == 'bind':
            Bind(responsible_process, port, tracker)
        elif operation == 'accept':
            Accept(responsible_process, the_parent_tracker_nbr(fields), writerp(fields), readerp(fields))
        else: # connects
            iden = 'port-' + id + '-from-' + responsible_process.name
            sock = resource_factory(PRGM, iden, 'Unix_Stream_Socket', 'self')
            sock.responsible_process = responsible_process
            if writerp(fields):
                responsible_process.add_transition('write', iden)
            if readerp(fields):
                responsible_process.add_transition('read', iden)
        #print sock
        #print fields[2] + "-->" + port + ":" + tracker + "+" + the_parent_tracker_nbr(fields)

class ReadDataFile:
        """A class for reading data files."""
        def __init__(self, filename, visitor, folder = ""):
            """ Initializes the class by opening the file
             """
            pathname = config.data_directory + "/" + folder + '/' + filename
            #print pathname
            #try:
            print "Reading from " + pathname
            self.file = open(pathname, 'r')
            self.first_line = 1
            self.name = filename
        def readlines (self):
            global resources_dict
            start_process = resources_dict[self.name[0:-8]]
            no_fork = 1
            for l in self.file.readlines():
                #valid lines come in several forms
                #file operations - pid|Open|filename|sc at file open time|file vs. directory|R or ""|sc of file read|W or ""|sc of file written to|fd|sc of invoking process
                #execs - pid|Exec|filename|sc of file|return code|sc of invoking process
                #clones - pid|Clone|pid of clone|sc of invoking process and clone
                #pipe - pid|Pipe| fd of complement pipe at creation time|R or ""|sc of file read|W or ""|sc of file written to| its fd|sc of process
                #stream - pid|Stream| ... ignore for now
                fields = l.split('\t')
                #print fields
                if len(fields) != 0:
                    #print fields
                    pid = the_pid(fields)
                    sc = the_process_sc(fields)
                    action = the_action(fields)
                    if self.first_line:
                         start_process.security_context = sc # so that application will always have a sc
                         start_process.original_context = sc
                         start_process.pid = pid
                         start_process.processes[pid] = start_process
                         self.first_line = 0
                    if pid in start_process.processes.keys():
                        responsible_process = start_process.processes[pid]
                        if responsible_process.security_context == 'unknown':
                            print "handling " + responsible_process.name
                            responsible_process.security_context = sc
                            responsible_process.original_context = sc
                    else:
                        print "pid problem"
                        print fields
                        responsible_process = processes[pid]
                    pipes = {}

                    obj = eval(action + 'Event()')
                    #print obj.__class__.__name__
                    obj.handle(start_process,responsible_process, fields)

        def close(self):
            self.file.close()

#################### accessors #############################

def the_process_sc (fields):
    return fields[-1][0:-1] # strip off \n

def the_pid(fields):
    return fields[0]

def the_action(fields):
    return fields[1]

def the_return_code(fields):
    return fields[4]

def succeeds (fields):
    if the_return_code(fields) == '-1':
        return 0
    else:
        return 1

def the_resource_name(fields):
    rname = fields[2]
    if rname[0] == '"':
        rname = rname[1:-1]
    return rname

def the_resource_sc(fields):
    """ for Open actions"""
    if the_action(fields) not in ['Open', 'Exec']:
        print "Inappropriate resource_sc query on " + str(fields)
        toreturn = 'unknown'
    toreturn = fields[3] # for execs
    if the_action(fields) != 'Exec':
        if writerp(fields):
            toreturn = fields[8]
        elif readerp(fields):
            toreturn = fields[6]
    return toreturn

def the_file_descriptor(fields):
    return fields[9]

def readerp (fields):
    return fields[5] == 'R'

def writerp (fields):
    return fields[7] == 'W'

def pipe_readerp (fields):
    return fields[3] == 'R'

def pipe_writerp (fields):
    return fields[5] == 'W'

def the_tracker_nbr (fields):
    return fields[-2]

def the_parent_tracker_nbr (fields):
    return fields[3]

def the_port(fields):
    toreturn = "-1000" #  fields[-3] # file descriptor
    start = fields[4].split(',')
    for s in start:
        pos = s.find('sin_port')
        if pos != -1:
            toreturn = s[pos + 15:-1]
            break
    return toreturn

#################### utilities ###############################
def remove_string(str, toremove):
    loc = str.find(toremove)
    if loc == -1:
        return str
    else: # its there
        return str[0:loc] + str[loc + len(toremove):]

def is_file (fields):
    if fields[4] == 'FILE':
        return 1
##    elif len(fields) == 4 and fields[2] in ['W', 'R']:
##        #old test
##        return 1
    else:
        return 0

def is_directory (fields):
    #temporary ad hoc
    #return fields[0][-1] == '/'
    return fields[4] == 'DIRECTORY'

def visit_files(prgm_name,folder):
    """Visit the structures for each file in folder"""
    #print "Reading from " + config.data_directory + "/" + folder
    base = config.data_directory + "/" + folder
    files = [x for x in os.listdir(base) if x.find('tracked') != -1]
    global PRGM
    PRGM = Program(prgm_name)
    #p = Process('startup', 'startup')
    #p.security_context = 'DEFAULT_t'
    for f in files:
        print f
        if stat.S_ISREG(os.stat(base + f)[0]): # test for file vs. directory
            p = resource_factory(PRGM, f[0:-8], 'Process', 'default_context')
            applications.append(p)
            r = ReadDataFile(f, None, folder)
            r.readlines()

def fix_accepts():
    for p in [v for v in resources_dict.values() if v.is_process()]:
        for a in p.with_accepts:
            port = a.listening_on() # port might be unknown
##            print a.process.name + ' listening on ' + port
            port_holders = [x for x in resources_dict.keys() if x.find('port-' + str(port)) != -1]
            for ph in port_holders:
                if a.writes:
                    p.add_transition('write', ph)
##                    print a.process.name + ' writing to ' + ph
                if a.reads:
                    p.add_transition('read', ph)
##                    print a.process.name + ' reading from ' + ph
